library pattern;

{version: 2010.2.13.1}
{$mode objfpc}{$H+}{$calling stdcall}

{   Copyright (c) 2010 Bernd Kreuss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program attached at the end of this source code file.
    If not, see <http://www.gnu.org/licenses/>.}

uses
  patternfunction;

const
  pattern_x_range = 100; // the domain of the pattern function pf()
  pattern_x_div   = 10;  // precision for determining the codomain of pf()
  scan_min_width  = 10;  // dont scan in less than this number of bars

  // indices used in the TPair arrays
  I_LOW    = 0;
  I_HIGH   = 1;
  I_WIDTH  = 0;
  I_ERROR  = 1;
  I_SCALE  = 0;
  I_OFFSET = 1;

type
  { The following data structure is an MT4 specific thing.
  Many thanks go to chipdude from the mql4.com forum
  for reverse engineering this part of the puzzle.
  The array itself is actually never instantiated by this dll,
  all we use is a pointer to such an array given to us
  by the calling mql4 program so we can peek around in
  metatrader's memory.}
  TCandle = packed record
    time:   integer;
    open,
    high,
    low,
    close,
    volume: double;
  end;
  THistory = array[0..65000] of TCandle;

  {in mql4 this is a double foo[2], used to pass a pair
  of doubles by reference from MT4 to an imported function}
  TPair = array[0..1] of double;


{ Calculate the range of close prices. The absolute values
of min and max will be written to the out variable price_range.}
procedure calcPriceRange(width, last_bar: integer; var history: THistory; bars: integer; out price_range: TPair);
var
  i:   integer;
  i0:  integer;
  i1:  integer;
  min: double;
  max: double;

begin
  i1  := bars - last_bar - 1;
  i0  := i1 - abs(Width);
  min := history[i0].Close;
  max := min;

  for i := i0 to i1 do
  begin
    if history[i].Close < min then min := history[i].Close;
    if history[i].Close > max then max := history[i].Close;
  end;

  price_range[I_LOW]  := min;
  price_range[I_HIGH] := max;
end;

{ Determine the y size of our pattern function.
The result is returned in the variable size. }
procedure patternSize(var par: TParam; out pattern_size: TPair);
var
  i:              integer;
  x, y, min, max: double;

begin
  max := pf(0, par);
  min := max;

  for i := 0 to pattern_x_range * pattern_x_div do
  begin
    x := i / pattern_x_div;
    y := pf(x, par);
    if y > max then max := y;
    if y < min then min := y;
  end;
  pattern_size[I_LOW]  := min;
  pattern_size[I_HIGH] := max;
end;

{ scale the pattern into the price rectangle defined by width
and last_bar. It uses the pattern height and the actual price
range to calculate offset and scale. You can then just use
scale[I_OFFSET] + scale[I_SCALE] * pf(x, p)
to determine the absolute prices.}
procedure scalePatternY(width: integer; var history: THistory; bars: integer; var par: TParam; var price_range, pattern_size: TPair; out scale: TPair);
var
  o: double; // price of pf(0) when scaled but not yet shifted to close(0)

begin
  scale[I_SCALE] := (price_range[I_HIGH] - price_range[I_LOW]) / (pattern_size[I_HIGH] - pattern_size[I_LOW]);

  if width > 0 then
  begin
    scale[I_OFFSET] := price_range[I_LOW] - pattern_size[I_LOW] * scale[I_SCALE];
    o               := scale[I_OFFSET] + scale[I_SCALE] * pf(0, par);
  end
  else
  begin
    // negative width means pattern will be inverted
    // so low becomes high and a few signs change
    scale[I_OFFSET] := price_range[I_LOW] + pattern_size[I_HIGH] * scale[I_SCALE];
    o               := scale[I_OFFSET] - scale[I_SCALE] * pf(0, par);
  end;

  // shift the pattern so that pf(0) always aligns with close(0)
  scale[I_OFFSET] += history[bars - 1].Close - o;
end;

{ Calculate the error value for a given price range. The pattern
will be scaled into the enclosing rectangle and the error between
pattern and actual close prices will be calculated and returned.}
function error(width, last_bar: integer; var history: THistory; bars: integer; var par: TParam; var price_range, pattern_size: TPair): double;
var
  i:      integer;
  err:    double;
  scale:  TPair;
  sign:   integer;
  end_i:  integer;
  fact_y: double;
  fact_x: double;

begin
  err  := 0;
  sign := 1;

  scalePatternY(width, history, bars, par, price_range, pattern_size, scale);

  if width < 0 then
  begin
    width := -width;
    sign  := -1;
  end;

  fact_y := sign * scale[I_SCALE];
  fact_x := pattern_x_range / width;
  end_i  := bars - last_bar - 1;

  for i := 0 to width do
  begin
    err += (abs(scale[I_OFFSET] + fact_y * pf(fact_x * i, par) - history[end_i - i].Close));
  end;

  // normalize it
  result := err / (width * (price_range[I_HIGH] - price_range[I_LOW]) * history[end_i].Close);
end;

{ Loop through all widths and find the one with the smallest error
range is the test range, all pattern widths up to range will be tested.
The best fitted width and its error are returned in the variable 'found' }
procedure findBest(range: integer; var history: THistory; bars: integer; var par: TParam; out found: TPair);
var
  i:            integer;
  e:            double;
  e_min:        double;
  r_best:       integer;
  price_range:  TPair;
  pattern_size: TPair;

begin
  e_min  := 999;
  r_best := scan_min_width;
  PatternSize(par, pattern_size);

  for i := scan_min_width to range do
  begin
    calcPriceRange(i, 0, history, bars, price_range);
    e := error(i, 0, history, bars, par, price_range, pattern_size);
    if e < e_min then
    begin
      e_min  := e;
      r_best := i;
    end;
    // inversed pattern (denoted by negative width)
    e := error(-i, 0, history, bars, par, price_range, pattern_size);
    if e < e_min then
    begin
      e_min  := e;
      r_best := -i;
    end;
  end;

  found[I_WIDTH] := r_best;
  found[I_ERROR] := e_min;
end;

exports
  pf,
  calcPriceRange,
  scalePatternY,
  patternSize,
  findBest;

end.

